/* Name: main.c
 * Project: AVR bootloader HID
 * Author: Christian Starkjohann
 * Creation Date: 2007-03-19
 * Tabsize: 4
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: Proprietary, free under certain conditions. See Documentation.
 * This Revision: $Id: main.c 373 2007-07-04 08:59:36Z cs $
BOOLEAN __stdcall   HidD_GetSerialNumberString(IN HANDLE device, OUT void *buffer, IN ULONG bufferLen);
 */

#include <stdio.h>
#include <windows.h>
//#include <string.h>
//#include <stdlib.h>
//#include <errno.h>
#include "usbcalls.h"

#define IDENT_VENDOR_NUM        0x1C40
#define IDENT_VENDOR_STRING     "MYDenshi"
#define IDENT_PRODUCT_NUM       0x0BBF
#define IDENT_PRODUCT_STRING    "HIDsph0"

/* ------------------------------------------------------------------------- */

usbDevice_t *dev = NULL;
static char dataBuffer[65536 + 256];    /* buffer for file data */
static int  startAddress, endAddress;
char optchr;

/* ------------------------------------------------------------------------- */

#define HID_R_FEATURE  USB_HID_REPORT_TYPE_FEATURE

static unsigned char resbuf[7];
static unsigned char cbuf[4];

typedef struct deviceInfo{
    char    reportId;
    char    pageSize[2];
    char    flashSize[4];
}deviceInfo_t;

static union{
    unsigned char   byt[1];
    deviceInfo_t    info;
}           buf;

int hidasp_init() {
    int len, err;

    err = usbOpenDevice(&dev, IDENT_VENDOR_NUM, IDENT_VENDOR_STRING,
                            IDENT_PRODUCT_NUM, IDENT_PRODUCT_STRING, 1);
    if(err) {
        fprintf(stderr, "ERROR: fail to usbOpenDevice()\n");
        return err;
    }
    Sleep(100);

    printf("HIDspx0 Connection check!");
    buf.byt[0] = 1;
    buf.byt[1] = 1; // Connection test
    buf.byt[2] = 0;
    while(buf.byt[2] < 4) {
        err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
        if(err) {
            fprintf(stderr, "\nERROR: fail to write()\n");
            return err;
        }
        Sleep(10);
        buf.byt[2]++;
    }

    len = sizeof(resbuf);
    err = usbGetReport(dev, HID_R_FEATURE, 1, resbuf, &len);
    if(err) return err;
    if(len < sizeof(resbuf)) err = -1;
    if(!err && (resbuf[1] != 3)) err = -2;
    if(err) {
        fprintf(stderr, "\nERROR: fail to read()\n");
        return err;
    }
    Sleep(10);
    printf(" OK.\n");
    return 0;
}

int hidasp_cmd(const unsigned char cmd[4], unsigned char res[4]) {
    int len, err;

    memcpy(buf.byt+2, cmd, 4);
    buf.byt[0] = 1;
    buf.byt[1] = 16; // SPI
    err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
    if(err) return err;
//  printf("hidasp_cmd %d cmd: %02x %02x %02x %02x ",buf[1],cmd[0],cmd[1],cmd[2],cmd[3]);

    if(res) {
        len = sizeof(resbuf);
        err = usbGetReport(dev, HID_R_FEATURE, 1, resbuf, &len);
        if(err) return err;
        if(len < sizeof(resbuf)) {
            fprintf(stderr, "ERROR: fail to command()\n");
            return -1;
        }
    memcpy(res, resbuf+1, 4);
//  printf("  res: %02x %02x %02x %02x\n",res[0],res[1],res[2],res[3]);
    }
//  printf("\n");
    return 0;
}

int hidasp_program_enable(int delay) {
    int err;

    printf("hidasp_program_enable()\n");

    Sleep(2);
    buf.byt[0] = 1;
    buf.byt[1] = 2;
    buf.byt[2] = 0x02; // LED off
    buf.byt[3] = 0x10; // ~RESET HI
    err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
    Sleep(10);
    if(err) return err;

    buf.byt[0] = 1;
    buf.byt[1] = 2;
    buf.byt[2] = 0x00; // LED on
    buf.byt[3] = 0x00; // ~RESET LOW
    err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
    if(err) return err;

    buf.byt[0] = 1;
    buf.byt[1] = 18; // SET_DELAY
    buf.byt[2] = delay; // *3inst. delay (min 6 : 1.5uSec.)
    buf.byt[3] = 0;
    err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
    if(err) return err;

    Sleep(60);
    cbuf[0]=0xAC; cbuf[1]=0x53; cbuf[2]=0x00; cbuf[3]=0x00;
    err = hidasp_cmd(cbuf,resbuf);
    if(err) return err;
    if(resbuf[2] != 0x53) {
        fprintf(stderr, "ERROR: no response from target device\n");
        return -3;
    }
    return 0;
}

void hidasp_close() {
    if(dev) {
        cbuf[0]=0x00; cbuf[1]=0x00; cbuf[2]=0x00; cbuf[3]=0x00;
        hidasp_cmd(cbuf, NULL); // AVOID BUG!
        buf.byt[0] = 1;
        buf.byt[1] = 2;
        buf.byt[2] = 0x02; // LED off
        buf.byt[3] = 0x10; // ~RESET HI
        usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
    }
    if(dev != NULL)
        usbCloseDevice(dev);
}

/* ------------------------------------------------------------------------- */

char    *usbErrorMessage(int errCode) {
static char buffer[80];

    switch(errCode){
        case USB_ERROR_ACCESS:      return "Access to device denied";
        case USB_ERROR_NOTFOUND:    return "The specified device was not found";
        case USB_ERROR_BUSY:        return "The device is used by another application";
        case USB_ERROR_IO:          return "Communication error with device";
        default:
            sprintf(buffer, "Unknown USB error %d", errCode);
            return buffer;
    }
    return NULL;    /* not reached */
}

#define InputReportByteLength  6
#define OutputReportByteLength 6
// char ttt=0;

int hidasp_page_write(long addr, const unsigned char *wd, int pagesize) {
    int err, n, l;

    // set page
    buf.byt[0] = 1;
    buf.byt[1] = 20; // Set Page mode
    buf.byt[2] = addr & 1 ? 0x48 : 0x40; // Flash
    buf.byt[3] = (unsigned char)((addr >> 9) & 0xFF);
    buf.byt[4] = (unsigned char)((addr >> 1) & 0xFF);
    err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
    if(err) {
        fprintf(stderr, "Error setup page address: %s\n", usbErrorMessage(err));
        return err;
    }

    // Load the page data into page buffer
    n = 0;
    while (n < pagesize) {
        l = OutputReportByteLength - 1; // MAX
        if(pagesize-n < l) l = pagesize - n;
        buf.byt[0] = 1;
        buf.byt[1] = 0x40 | l; // PageBuf
        memcpy(buf.byt+2, wd+n, l);
//      Sleep(2);
/*      if(!ttt++) {
            printf("  p: %02x %02x %02x %02x %02x\n",buf.byt[2],buf.byt[3],
                                          buf.byt[4],buf.byt[5],buf.byt[6]);} */
        err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
        if(err) {
            fprintf(stderr, "Error uploading data block: %s\n", usbErrorMessage(err));
            return err;
        }
//      printf("  p: %02x %02x %02x %02x\n",buf[1],buf[2],buf[3],buf[4]);
        n += l;
    }
    return 0;
}

int hidasp_page_read(long addr,unsigned char *wd,int pagesize) {
    int err, n, l, len;

    // set page
    if(addr >= 0) {
        buf.byt[0] = 1;
        buf.byt[1] = 20; // Set Page mode
        buf.byt[2] = addr & 1 ? 0x28 : 0x20; // Flash read
        buf.byt[3] = (unsigned char)((addr >> 9) & 0xFF);
        buf.byt[4] = (unsigned char)((addr >> 1) & 0xFF);
        err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
        if(err) {
            fprintf(stderr, "Error setup page address: %s\n", usbErrorMessage(err));
            return err;
        }
    }

    // Load the page data into page buffer
    n = 0;
    while (n < pagesize) {
        l = OutputReportByteLength - 1; // MAX
        if(pagesize - n < l) l = pagesize - n;
        buf.byt[0] = 1;
        buf.byt[1] = 0x40 | l; // PageBuf
        memset(buf.byt+2, 0, l);
//      Sleep(2);
        err = usbSetReport(dev, HID_R_FEATURE, buf.byt, sizeof(buf.info));
        if(err) {
            fprintf(stderr, "Error reading data block(command): %s\n",
                                                 usbErrorMessage(err));
            return err;
        }
        len = sizeof(resbuf);
        err = usbGetReport(dev, HID_R_FEATURE, 1, resbuf, &len);
        if(!err && (len < sizeof(resbuf)) ) err = -1;
        if(err) {
            fprintf(stderr, "Error reading data block: %s\n", usbErrorMessage(err));
            return err;
        }
        memcpy(wd+n, resbuf+1, l);
//      printf("  p: %02x %02x %02x %02x\n",buf[1],buf[2],buf[3],buf[4]);
        n += l;
    }
    return 0;
}

/* ------------------------------------------------------------------------- */

static int uploadData(unsigned char *dataBuffer, int startAddr, int endAddr) {
int         err = 0, mask, pageSize, deviceSize, startAddrV, endAddrV;
unsigned char buffer[16384];

    err = hidasp_init();
    if(err) {
        fprintf(stderr, "Error opening HIDspx0 device: %s\n", usbErrorMessage(err));
        goto errorOccurred;
    }

    err = hidasp_program_enable((int)12);
    if(err) {
        fprintf(stderr, "Error opening HIDspx device: %s\n", usbErrorMessage(err));
        goto errorOccurred;
    }

    pageSize = 32; // getUsbInt(buffer.info.pageSize, 2);
    deviceSize = 0x800; // getUsbInt(buffer.info.flashSize, 4);
    printf("\nPage size   = %d (0x%X)\n", pageSize, pageSize);
    printf("Device size = %d (0x%X)\n", deviceSize, deviceSize);
    if(endAddr > deviceSize) {
        fprintf(stderr, "Data (%d bytes) exceeds remaining flash size!\n", endAddr);
        err = -1;
        goto errorOccurred;
    }
    mask = pageSize - 1;
    startAddr &= ~mask;                  /* round down */
    endAddr = (endAddr + mask) & ~mask;  /* round up */
    startAddrV = startAddr;
    endAddrV = endAddr;

    if(optchr != 'V') {
        // chip erase
        cbuf[0] = 0xAC; cbuf[1] = 0x80; cbuf[2] = 0; cbuf[3] = 0;
        err = hidasp_cmd(cbuf, NULL);
        Sleep(20);
        if(err) {
            fprintf(stderr, "Error device erassing: %s\n", usbErrorMessage(err));
            goto errorOccurred;
        }

        printf("Uploading %d (0x%x) bytes starting at %d (0x%x)\n", endAddr - startAddr,
                                              endAddr - startAddr, startAddr, startAddr);
        while(startAddr < endAddr) {
            memcpy(buffer, dataBuffer + startAddr, pageSize);
            printf("\r0x%05x ... 0x%05x", startAddr, startAddr + pageSize);
            fflush(stdout);
            err = hidasp_page_write((long)startAddr, buffer, (int)pageSize);
            if(err) goto errorOccurred;

            // writting
            cbuf[0] = 0x4C; // Flash write
            cbuf[1] = (unsigned char)((startAddr >> 9) & 0xFF);
            cbuf[2] = (unsigned char)((startAddr >> 1) & 0xFF);
            cbuf[3] = 0;
            err = hidasp_cmd(cbuf, NULL);
            Sleep(20);
            if(err) {
                fprintf(stderr, "Error writting page: %s\n", usbErrorMessage(err));
                goto errorOccurred;
            }
            startAddr += pageSize;
        }
        printf("\n");
    }

    if(optchr != 'W') { int i;
        printf("Verfy     %d (0x%X) bytes starting at %d (0x%X)\n", endAddrV - startAddrV,
                                            endAddrV - startAddrV, startAddrV, startAddrV);
        while(startAddrV < endAddrV) {
            printf("\r0x%05x ... 0x%05x", startAddrV, startAddrV + pageSize);
            fflush(stdout);
            err = hidasp_page_read((long)startAddrV, buffer, (int)pageSize);
            if(err) goto errorOccurred;
//          if(startAddrV >= 0x600) buffer[30] = ~buffer[30];
            for(i = 0; i < pageSize; ++i) {
                if(buffer[i] != dataBuffer[startAddrV + i]) {
                    printf("\nVerfy error %d (0x%X)=0x%02X:0x%02X\n", startAddrV + i, startAddrV + i,
                                             dataBuffer[startAddrV + i], buffer[i]);
                    goto errorOccurred;
                }
            }
            startAddrV += pageSize;
        }
        printf("\n");
    }

errorOccurred:
    hidasp_close();
    return err;
}

/* ------------------------------------------------------------------------- */

int readfuse(unsigned char *fusel, unsigned char *fuseh, unsigned char *fusee) {
    int err;

    cbuf[0]=0x50; cbuf[1]=0x00; cbuf[2]=0x00; cbuf[3]=0x00;
    err = hidasp_cmd(cbuf,resbuf);
    if(err) return err;
    *fusel = resbuf[4];
    cbuf[0]=0x58; cbuf[1]=0x08; cbuf[2]=0x00; cbuf[3]=0x00;
    err = hidasp_cmd(cbuf,resbuf);
    if(err) return err;
    *fuseh = resbuf[4];
    cbuf[0]=0x50; cbuf[1]=0x08; cbuf[2]=0x00; cbuf[3]=0x00;
    err = hidasp_cmd(cbuf,resbuf);
    if(err) return err;
    *fusee = resbuf[4];
    return 0;
}

int writefuse(unsigned char fusel, unsigned char fuseh, unsigned char fusee) {
    int err;

    cbuf[0]=0xAC; cbuf[1]=0xA0; cbuf[2]=0x00; cbuf[3]=fusel;
    err = hidasp_cmd(cbuf, NULL);
    Sleep(20);
    if(err) return err;
    cbuf[0]=0xAC; cbuf[1]=0xA8; cbuf[2]=0x00; cbuf[3]=fuseh;
    err = hidasp_cmd(cbuf, NULL);
    Sleep(20);
    if(err) return err;
    cbuf[0]=0xAC; cbuf[1]=0xA4; cbuf[2]=0x00; cbuf[3]=fusee;
    err = hidasp_cmd(cbuf, NULL);
    Sleep(20);
    if(err) return err;
    return 0;
}

char bitstr[9];
char *bstr(char a) {
unsigned char k;
int i;
    for(k = 0x80, i = 0; i < 8; ++i, k >>= 1)
        bitstr[i] = ((a & k) ? '1' : '0');
    bitstr[8] = 0;
    return bitstr;
}

int fuse(unsigned int *fsl, unsigned int *fsh, unsigned int *fse) {
    unsigned char fusel, fuseh, fusee, ch;
    int err;

    err = hidasp_init();
    if(err) {
        fprintf(stderr, "Error opening HIDspx0 device: %s\n", usbErrorMessage(err));
        goto error;
    }
    err = hidasp_program_enable((int)12);
    if(err) {
        fprintf(stderr, "Error opening HIDspx device: %s\n", usbErrorMessage(err));
        goto error;
    }
    err = readfuse(&fusel, &fuseh, &fusee);
    if(err) goto error;
    printf("L:%02X(%8s)", (int)fusel, bstr(fusel));
    printf(" H:%02X(%8s)", (int)fuseh, bstr(fuseh));
    printf(" E:%02X(%8s) before\n", (int)fusee, bstr(fusee));
    if(!(*fsl | *fsh | *fse) ) goto error;
    fusel = (unsigned char)(*fsl & 0xFF);
    fuseh = (unsigned char)(*fsh & 0xFF);
    fusee = (unsigned char)(*fse & 0xFF);
    printf("L:%02X(%8s)", (int)fusel, bstr(fusel));
    printf(" H:%02X(%8s)", (int)fuseh, bstr(fuseh));
    printf(" E:%02X(%8s) to write\ntype 'Y' to write ? ", (int)fusee, bstr(fusee));
    ch = getch();
    if(toupper(ch) == 'Y') {
        err = writefuse(fusel, fuseh, fusee);
        if(err) goto error;
        err = readfuse(&fusel, &fuseh, &fusee);
        if(err) goto error;
        printf("\rL:%02X(%8s)", (int)fusel, bstr(fusel));
        printf(" H:%02X(%8s)", (int)fuseh, bstr(fuseh));
        printf(" E:%02X(%8s) writen\n", (int)fusee, bstr(fusee));
    }else printf("\n");
error:
    hidasp_close();
    return err;
}

/* ------------------------------------------------------------------------- */

static int  parseUntilColon(FILE *fp) {
int c;

    do{
        c = getc(fp);
    }while(c != ':' && c != EOF);
    return c;
}

static int  parseHex(FILE *fp, int numDigits) {
int     i;
char    temp[9];

    for(i = 0; i < numDigits; i++)
        temp[i] = getc(fp);
    temp[i] = 0;
    return strtol(temp, NULL, 16);
}

/* ------------------------------------------------------------------------- */

static int  parseIntelHex(char *hexfile, char buffer[65536 + 256], int *startAddr, int *endAddr) {
int     address, base, d, segment, i, lineLen, sum;
FILE    *input;

    input = fopen(hexfile, "r");
    if(input == NULL){
        fprintf(stderr, "error opening %s: %s\n", hexfile, strerror(errno));
        return 1;
    }
    while(parseUntilColon(input) == ':'){
        sum = 0;
        sum += lineLen = parseHex(input, 2);
        base = address = parseHex(input, 4);
        sum += address >> 8;
        sum += address;
        sum += segment = parseHex(input, 2);  /* segment value? */
        if(segment != 0)    /* ignore lines where this byte is not 0 */
            continue;
        for(i = 0; i < lineLen ; i++){
            d = parseHex(input, 2);
            buffer[address++] = d;
            sum += d;
        }
        sum += parseHex(input, 2);
        if((sum & 0xff) != 0){
            fprintf(stderr, "Warning: Checksum error between address 0x%x and 0x%x\n", base, address);
        }
        if(*startAddr > base)
            *startAddr = base;
        if(*endAddr < address)
            *endAddr = address;
    }
    fclose(input);
    return 0;
}

/* ------------------------------------------------------------------------- */

static void printUsage(char *pname) {
    fprintf(stderr, "usage: %s [-v][-w] <intel-hexfile> ;present device is only ATTiny2313\n", pname);
    fprintf(stderr, "       %s -f [64DF01] ;fuse LOW:HI:EXT\n", pname);
    fprintf(stderr, "      CR 1MHz 64\n");
    fprintf(stderr, "      CR 8MHz E4\n");
    fprintf(stderr, "      OCLOCK   E0\n");
    fprintf(stderr, "      XTAL  ^8MHz FF\n");
    fprintf(stderr, "      ZbN  CF\n");
    fprintf(stderr, " HIDSPH(Tiny2313) 8F\n");
}

int main(int argc, char **argv) {
char    *file;
int fusel, fuseh, fusee;

    if(argc < 2) {
        printUsage(argv[0]);
        return 1;
    }
    if(argv[1][0] == '-') {
        optchr = toupper(argv[1][1]);
        if(optchr == 'H') {
            printUsage(argv[0]);
            return 1;
        }
        if(optchr =='F') {
            int i; char ch;
            fusel = 0; fuseh = 0; fusee =0;
            if(argc < 3) goto fus;
            for(i = 0; i < 6; ++i) {
                ch = toupper(argv[2][i]);
                if( !((ch >= '0') && (ch <= '9')) && !((ch >= 'A') && (ch <= 'F')) )
                    goto fus;
            }
            sscanf( argv[2],   "%2x", &fusel);
            sscanf(&argv[2][2],"%2x", &fuseh);
            sscanf(&argv[2][4],"%2x", &fusee);
fus:
            fuse(&fusel, &fuseh, &fusee);
            return 0;
        }

        if((argc < 3) || ((optchr != 'V') && (optchr != 'W')) ) {
            printUsage(argv[0]);
            return 1;
        }
        file = argv[2];
    }else {
        file = argv[1];
    }
    startAddress = sizeof(dataBuffer);
    endAddress = 0;
    memset(dataBuffer, -1, sizeof(dataBuffer));
    if(parseIntelHex(file, dataBuffer, &startAddress, &endAddress) )
        return 1;
    if(startAddress >= endAddress) {
        fprintf(stderr, "No data in input file, exiting.\n");
        return 0;
    }
    if(uploadData(dataBuffer, startAddress, endAddress) )
        return 1;
    return 0;
}

/* ------------------------------------------------------------------------- */
