/* Name: main.c
 * Project: AVR bootloader HID
 * Author: Christian Starkjohann
 * Creation Date: 2007-03-19
 * Tabsize: 4
 * Copyright: (c) 2007 by OBJECTIVE DEVELOPMENT Software GmbH
 * License: GNU GPL v2 (see License.txt)
 * This Revision: $Id: main.c 378 2007-07-07 20:47:21Z cs $
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>
#include <string.h>

#include "bootloaderconfig.h"
#include "usbdrv.c"

static char             reportId = -1;
// static unsigned long    currentAddress; /* in bytes */
static uchar            offset;         /* data already processed in current transfer */

static uchar    usbData[8];
static uchar page_mode;

typedef struct shortaddr{
    uchar   lbyte;
    uchar   hbyte;
}shrtaddr_t;

union{
    uchar       bytes[1];
    uint16_t    w;
    shrtaddr_t  b;
}               page_addr;

PROGMEM char usbHidReportDescriptor[42] = {
    0x06, 0x00, 0xff,              // USAGE_PAGE (Generic Desktop)
    0x09, 0x01,                    // USAGE (Vendor Usage 1)
    0xa1, 0x01,                    // COLLECTION (Application)
    0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
    0x26, 0xff, 0x00,              //   LOGICAL_MAXIMUM (255)
    0x75, 0x08,                    //   REPORT_SIZE (8)

    0x85, 0x01,                    //   REPORT_ID (1)
    0x95,    6,                    //   REPORT_COUNT (6)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)

    0x85, 0x02,                    //   REPORT_ID (2)
    0x95,    7,                    //   REPORT_COUNT (7)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)

    0x85, 0x03,                    //   REPORT_ID (3)
    0x95,   38,                    //   REPORT_COUNT (38)
    0x09, 0x00,                    //   USAGE (Undefined)
    0xb2, 0x02, 0x01,              //   FEATURE (Data,Var,Abs,Buf)

    0xc0                           // END_COLLECTION
};


static uchar wait=60; // 160
static void delay(uchar d) {
	while(d--);
}

static uint8_t usi_trans(uchar data) {
	USIDR = data;
	USISR = (1 << USIOIF);
	do{
		delay(wait);
		USICR = (1 << USIWM0) | (1 << USICS1) | (1 << USICLK) | (1 << USITC);
	} while(!(USISR & (1 << USIOIF)) );
	return USIDR;
}

uchar   usbFunctionSetup(uchar data[8]) {
usbRequest_t    *rq = (void *)data;

    if(rq->bRequest == USBRQ_HID_SET_REPORT){
        reportId = rq->wValue.bytes[0];    /* store report ID */
        offset = 0;
        return 0xff;
    }else if(rq->bRequest == USBRQ_HID_GET_REPORT){
        usbData[0] = 1;                    /* report ID */
        usbMsgPtr = usbData;
        return 7;
    }
    return 0;
}

uchar usbFunctionWrite(uchar *data, uchar len) {
    if(reportId == 1) {
//  if(data[1] == 1) // TEST
        usbData[1] = data[2];
        if(data[1] == 2) { // SET_LED
//          PORTD = (PORTD & ~3) | (data[2] & 3);
//          PORTB = (PORTB & ~0x10) | (data[3] & 0x10);
//          PORTB = (PORTB & ~0x0C) | ((data[2] << 2) & 0x0C);
            PORTB = (PORTB & ~0x1C) | ((data[3] & 0x10) | ((data[2] << 2) & 0x0C));
        }else if(data[1] == 18) { // Set wait
            wait = data[2];
        }else if(data[1] == 16) { // SPI
            for(uchar i = 0; i < 4; i++) {
                usbData[i+1] = usi_trans(data[i+2]);
            }
        }else if(data[1] == 20) { // Page set
            page_mode = data[2];
            page_addr.b.hbyte = data[3];
            page_addr.b.lbyte = data[4];
        }else if( ((data[1] ^= 0x40) & 0xC0) == 0) { // Page buf
            for(uchar i = 0; i < data[1]; i++) {
                usi_trans(page_mode);
                usi_trans(page_addr.b.hbyte);
                usi_trans(page_addr.b.lbyte);
                usbData[i+1] = usi_trans(data[i+2]);
                if(page_mode & 0x88) { // EEPROM or Flash(P=Hi)
                    page_addr.w++;
                    page_mode &= ~0x08;
                }else {
                    page_mode |= 0x08;
                }
            }
        }
    }
    return 1;
}

int main(void) {
    PORTD = 0x00;
    DDRD = ~USBMASK;   /* all outputs except USB data */
    PORTB = 0x08;      /* no pullups on USB pins */ // LED off
    DDRB = ~(1<<5);    /* all outputs except USI input data */
    /* initialize hardware */
//  bootLoaderInit();

    usbInit();
    sei();
    while(1){ /* main event loop */
        usbPoll();
    }
}
